// Cecible_ESP32_v1.24                           YLC  20/11/21  
// version avec balayage restreint à la surface du carré circonscrit à la cible.
// id est : mise en oeuvre de l'amélioration décrite ci-dessous
//
// A compiler avec type de carte "DOIT ESP32 DEVKIT V1"  ou "ESP32 Dev Module" 
// NB : Les joueurs se positionnent dans l'axe de la diagonale allant du Lidar à l'angle opposé.   
// 
//
//                                     ▼ <--- Lidar
//                                   / | \
//          Côté de référence ---> /   |   \  <---  Côté de référence utilisateur
//          trigonométrique      /     |     \
//                              <--- Cible --->
//                               \     |     /
//                                 \   |   /
//                                   \ | /
//                                     V
//
//                                  Joueurs   
//  
// Version avec calage automatique du lidar (angle de calage fourni par utilitaire de calage à exécuter après construction)
// Le côté de référence utilisateurs est le côté en haut à droite de la cible vue depuis les joueurs.
// Le lidar effectue des balayage de la cible dans le sens horaire à partir de sa position de calage.
// La position de calage du lidar permet de démarrer les balayages avant le côté de référence utilisateur ...
// ... et de terminer après le côté de référence trigonométrique (pour détecter la balle si elle déborde en limite de cible)
// ...donc l'angle maximal de balayage possible est un peu supérieur à 90° 
// Chaque mesure de distance et en fait la moyenne de 5 mesures valides afin de lisser d'éventuelles erreurs de détection
// L'écran Oled est utilisé pour des tests éventuels et la mise au point du programme. Son activation se fait en modifiant le paramètre "OLED".
//
// Amélioration de cette version :
//     On limite la détection à la stricte surface du carré circonscrit à la cible et on s'affranchit des
//     obstacles qui entoureraient celle-ci (pied de spectateurs p.ex.) en asservissant la distance maxi de validité de détection 
//     proportionnellement à l'angle du lidar au moment de chaque mesure.
//     Cela implique la suppression du balayage rapide de l'ancienne version pour ne faire qu'un balayage lent afin de permettre de calculer, à chaque
//     progression de l'angle du lidar d'un pas, l'augmentation ou la diminution la distance maxi d'une valeur proportionnele au nombres de pas par rapport
//     au pas de calage et à celui correspondant à la diagonale du carré circonscrit à la cible. (Augmentation jusqu'à la diagonale puis diminution)
//    
// Composition : ESP32 Wroom 32 + Buzzer + Lidar tf mini plus  + moteur pàp 28BYJ-48 + contrôleur ULN2003A
// Hardware    :                + relais reed  + 2x switchs fin de course (fermés au repos) + détecteur optique à fourche
//             :                + convertisseur de niveau RX/TX <> 3v/5v + Oled 128x64 (pour mise au point) 
// ---------------------------------------------------------------------------------------
// Pour fonctionner correctement il faut au préalable initialiser les bonnes constantes suivantes dans le programme :
const float xmax = 100;      // Taille maxi en x du carré circonscrit à la cible 
const float ymax = 100;      // Taille maxi en y du carré circonscrit à la cible  
const float xlidar = 0;      // position en x du lidar  
const float ylidar = 0;      // position en y du lidar
const int diabal = 9;        // diamètre moyen de la balle 
const float margedeg = 10;   // marge angulaire de balayage (en degrés) (10 = approx. 50 pas)
const float angdep = 4.39;   // angle de calage (en degrés) fourni par l'utilitaire de mesure de cet angle (mesuré le 09/04/23)
const int ptsanno1 = 10;     // nb de point du premier anneau 
const int ptsanno2 = 20;     // nb de point du second anneau 
const int ptsanno3 = 30;     // nb de point du 3ème anneau
const int ptscentre = 50;    // nb de point du centre de la cible  
// ---------------------------------------------------------------------------------------
// Récapitulatif des assignation des broches de l'ESP32 :
// pin 2   Builtin-led   (virtuelle)
// pin 13  Buzzer
// pin 12  Détecteur optique à fourche
// pin 27  Relais reed de controle de fixation de la cible
// pin 26  Commande moteur #4
// pin 25  Commande moteur #3
// pin 33  Commande moteur #2
// pin 32  Commande moteur #1
// pin RX2 TX du Lidar
// pin TX2 RX du Lidar
// pin 18  Fin de course gauche
// pin 19  Fin de course droite
// pin 21  SDA écran Oled
// pin 22  SCL écran Oled
// ---------------------------------------------------------------------------------------
// écran Oled pour suivi de la mise au point si modifs à faire (optionnel)
#define OLED 0               // 1 pour activer le suivi de test sur écran Oled, 0 pour le désactiver
#define MOSER 0              // 1 pour activer le suivi de test sur le moniteur série, 0 pour le désactiver
#include <U8g2lib.h>         // librairie de polices pour écran Oled 
U8G2_SSD1306_128X64_NONAME_F_HW_I2C u8g2(U8G2_R0);  
// ---------------------------------------------------------------------------------------

// contrôle du calibrage
#define ledctl  2      // builtin led pour contrôle de calage du lidar (allumé si pas calé)

// buzzer
#include "EasyBuzzer.h"       // Attention : utilise le channel ledcpwm 0

// fixation de la cible sur le boitier
const int fixcib = 27;         // détecteur de fixation de la cible connecté sur pin 27

// module lidar TF mini plus  sur port série #2 (pins 16 et 17)
#include <HardwareSerial.h>
HardwareSerial lidar(2);
int16_t tfDist = -1;    // Distance to object in centimeters
int dis = 0;            // distance brute mesurée
int dist = 0;           // distance mesurée valide
int dislid = 0;         // cumul des mesures faite par le lidar 
const int HEADER = 0x59;    // identifiant des données lidar

#include "BluetoothSerial.h"
BluetoothSerial SerialBT;
String cmd="";

// moteur
#include <TinyStepper_28BYJ_48.h>
const int mot1 = 32;        // assignations des pins de commmande moteur  pin 32
const int mot2 = 33;        // pin 33
const int mot3 = 25;        // pin 25
const int mot4 = 26;        // pin 26  
TinyStepper_28BYJ_48 moteur;       // instance moteur (2048 pas par tour)
#define PI 3.1415926535897932384626433832795
long step90 = 512;      // nb steps pour balayer 90° (moteur fait 2048 pas / tour)
float radstep;          // nombre de radians par step
long angle;             // angle moyen de la position de l'obstacle (en pas) (vers centre de l'obstacle)
long margestep;         // Nombre de pas de balayage avant et après la cible 
long stepdeb;           // pas de début de balayage pour détection fine de la balle
long stepfin;           // pas de fin de balayage pour détection de la balle 
long stepref = 512;     // pas du côté de référence du cercle trigonométrique (coté opposé aux joueurs) (90° % coté droit)
long stepcal;           // pas de la position de calibrage par rapport au coté droit de la cible vue des joueurs (négatif)
float vadimax;          // pas de variation de la distance maxi en fonction du pas de l'angle du lidar    

// détecteurs de calage moteur
const int fdcg = 18;     // fin de course gauche sur pin 18
const int fdcd = 19;     // fin de course droite sur pin 19
const int lopt = 12;     // capteur optique de calage sur pin 12

// autres variables
long i,j,k,l,m = 0;
bool coderr = false;      // indicateur d'erreur
float distmin;            // distance minimale de détection de la balle
float distmax;            // distance maximale de détection de la balle
float largano;            // largeur des 4 anneaux colorés de la cible  
float xcentre;            // position en x du centre de la cible
float ycentre;            // position en y du centre de la cible
float xbal;               // position en x du centre de la balle
float ybal;               // position en y du centre de la balle
float discenbal;          // distance jusqu'au centre de la balle
char dg = ' ';            // position à droite ou à gauche par rapport au centre de la cible
char hb = ' ';            // position en haut ou en bas par rapport au centre de la cible
float discent;            // distance du centre de la balle par rapport au centre de la cible
float travx;              // pour calcul en x
float travy;              // pour calcul en y 
int point;                // point marqué (0 si hors cible, 1 pour anneau ext. 2,3, et 4 pour centre de cible)
unsigned char uart[9];    // données lues par le lidar

//__________________________________________________________________

void sonner(int freq, int duree) {
  EasyBuzzer.singleBeep(freq,duree);      // faire sonner la cible
  delay(duree);
  EasyBuzzer.stopBeep();
}
//__________________________________________________________________

void calibrage() {
  // NB : on tient compte qu'une saleté pourrait occulter le détecteur optique et en fausser sa lecture
  // d'où les tests des fins de course qui ne devraient normalement pas être sollicités.
  coderr = false;
  while (digitalRead(lopt) == LOW) {  // si le détecteur optique est occulté mais peut être mal calé 
    ledcWrite(0, 0);                  // éteindre la led de controle 
    moteur.moveRelativeInSteps(1);    // touner le lidar en sens horaire de 1 pas
   if (digitalRead(fdcd) == HIGH) {   // Si fin de course droit détectée (ne devrait pas arriver)
      moteur.setupStop();             // arrêt du balayage rapide 
      coderr = true; 
      break;
    }  
  }
  coderr = false;
  while (digitalRead(lopt) == HIGH) {  // tant que détecteur optique n'est pas occulté (donc à 5V)
    ledcWrite(0, 255);          // allumer la led de controle (la led s'éteindra dès que le lidar est calé)
    moteur.moveRelativeInSteps(-1);    // pour faire balayage anti-horaire de 1 pas
    if (digitalRead(fdcg) == HIGH) {   // Si fin de course gauche détectée => anomalie de fonctionnement
      coderr = true;
      break;   
    }  
  }
  ledcWrite(0, 0);              // éteindre la led de controle, la lidar est bien calé   
  moteur.setCurrentPositionInSteps(stepcal);  // attribuer la position de calage au moteur
  moteur.disableMotor();             // arrêt du moteur    
  stepfin = stepref + margestep;     // limite de fin de balayage pour détection de la balle
  #if OLED 
    u8g2.clearBuffer();
    u8g2.setCursor(0,10); ;  
    u8g2.print("fin= "); 
    u8g2.print(stepfin);  
    u8g2.setCursor(0,22);; 
    u8g2.print("marge=");
    u8g2.print(margestep);  
    u8g2.setCursor(0,64); 
    u8g2.print("Erreur=");
    u8g2.print(coderr); 
    u8g2.sendBuffer(); 
  #endif 
  #if MOSER 
    Serial.print("fin= "); 
    Serial.print(stepfin);  
    Serial.print("   marge=");
    Serial.print(margestep);  
    Serial.print("   Erreur=");
    Serial.println(coderr); 
  #endif
}

//__________________________________________________________________
 
void lectdist() {           // lecture distance lidar sans passer par bibliothèque
  tfDist = -1;             // lecture du lidar à faux par défaut
    
  if (lidar.available()) {         //cntrl si port série #2 a des données en entrée
    for (int w = 0; w < 9; w++) {    
      uart[w] = lidar.read();
    }
    if ((uart[0] == HEADER) && (uart[1] == HEADER)
    && ((unsigned char)(uart[0] + uart[1] + uart[2] + uart[3] + uart[4] + uart[5] + uart[6] + uart[7]) == uart[8])) {
      tfDist = uart[2] + uart[3] * 256;    //calculer la distance 
    } 
  }
  
  #if MOSER 
    Serial.print("   tfDist= ");
    Serial.print(tfDist);
  #endif 
} 
    
//__________________________________________________________________

void mesure()  {
  dis = dist;            // pour ne pas prendre en compte la distance si lecture non valide
  dislid = -1;
  m = 0;
  for (l = 0; l < 9; l++) {                      // on fait 9 mesures consécutives      
    lectdist();                                // Récuperation de la distance mesurée en cm    
    if ((tfDist > distmin) && (tfDist < distmax)) {  // si mesure valide  (sinon mesure inchangée) 
      if (dislid == -1) dislid = tfDist;
      else dislid = dislid + tfDist;   
      m++;  
    }     
  }  
  if (dislid > -1) {                        // si mesures valides effectuée
    dis = round(dislid / m);               // distance moyenne de 5 mesures                            
  } 
  #if MOSER 
    Serial.print("   dislid= ");
    Serial.println(dislid);
  #endif 
}
//__________________________________________________________________

void calcdist() {  
  if (dis < dist) {              //  Si distance diminue (centrage sur l'obstacle)
    sonner((2000 - (dist * 4)),50);  // détection sonore de l'obstacle                                  TEST a rétablir
    dist = dis;
    angle = j;                    // angle de détection la plus courte (en pas)
  }  
  if (dis > dist) {              // dès que la distance remonte => fin de détection
    j = stepfin;                  // pour sortir de la boucle
  }        
}
//__________________________________________________________________

 void detectballe() { 
  coderr = false;
  dist = 9999;                              // init dist stockée
  moteur.moveToPositionInSteps(stepcal);    // retour lidar en position de calage   
  delay(100);  
  distmax = xmax;                       // init dist maxi à longueur du côté
  for (j = stepcal; j < stepfin; j++) { // balayage pour détection lente
    moteur.moveToPositionInSteps(j);   
    if ((j > 0) && (j <= 255)) {        //  à chaque pas d'angle lidar positif en-deçà de la diagonale (255) ...
      distmax = distmax + vadimax;      //  ... on allonge la dist. maxi jusqu'à la valeur de la diagonale 
    }
    else if (j > 255) {                  //  à chaque pas d'angle lidar positif au-delà de la diagonale ...
      distmax = distmax - vadimax;     //  ... on réduit la dist. mini jusqu'à la longueur du côté
    }
    mesure();    
    if (dislid > -1) calcdist();       // cherche la distance minimale de la balle    
  }
  moteur.moveToPositionInSteps(stepcal);  // retour lidar en position de calage    
  moteur.disableMotor();                  //arrêt du moteur     
  dg = ' ';                               // init variables de restitution
  hb = ' ';
  point = 0;
  if (dist < 9999) {                  // si la balle a été détectée
    discenbal = dist + (diabal / 2);  // calcul distance jusqu'au centre de la balle
    xbal = (cos((stepref - angle) * radstep) * discenbal);  
    xbal = xbal - abs(xlidar);        // correction en x par rapport à la position du lidar
    ybal = (sin((stepref - angle) * radstep) * discenbal);
    ybal = ybal - abs(ylidar);        // correction en y par rapport à la position du lidar
//  Cible en diagonale par rapport aux joueurs => positionnement par rapport aux diagonales
    if (xbal > ybal)       dg = 'G';
    if (xbal < ybal)       dg = 'D';
    if (abs(xbal + ybal) > xmax)  hb = 'B';
    if (abs(xbal + ybal) < xmax)  hb = 'H';
    travx = pow(abs(xcentre - xbal), 2);   // calcul de la distance entre la balle et le centre de la cible
    travy = pow(abs(ycentre - ybal), 2);
    discent = sqrt(travx + travy);         
    if (discent <= largano) {         // si distance de la balle à la cible
      point = ptscentre;              // dans le milieu de la cible
      dg = 'C';                       // ie : "au centre"
      hb = ' ';
    }
    else if (discent <= (largano * 2))            point = ptsanno3;  // ...sinon sur troisième anneau 
         else if (discent <= (largano * 3))       point = ptsanno2;  // ...sinon sur second anneau
              else if (discent <= (largano * 4))  point = ptsanno1;  // ...sinon sur quatrième anneau                
  }
  
  lidar.end();                                // fermer port série
  delay(50);
  lidar.begin (115200,SERIAL_8N1,16,17) ;   // réouvre port série 2: Lidar sur pins 16 et 17
  
  #if MOSER 
    Serial.print("   stepref=");
    Serial.print(stepref); 
    Serial.print("   angle (pas)=");
    Serial.print(angle); 
    Serial.print("   angle (deg)=");
    float angdeg = (stepref - angle) * 0.17578125;     // 360 / 2048
    Serial.print(angdeg,2); 
    Serial.print("   radstep=");
    Serial.print(radstep,8); 
    Serial.print("   discenbal=");
    Serial.print(discenbal); 
    Serial.print("   stepcal=");
    Serial.print(stepcal);
    Serial.print("   stepfin=");
    Serial.print(stepfin); 
    Serial.print("   dist=");
    Serial.print(dist);  
    Serial.print("   travx=");
    Serial.print(travx); 
    Serial.print("   travy=");
    Serial.print(travy);
    Serial.print("   discent= ");
    Serial.print(discent);
    Serial.print("   xbal= ");
    Serial.print(xbal); 
    Serial.print("   point= ");
    Serial.print(point); 
    Serial.println("------");
  #endif   
  
  // si pas de détection on ne renvoie que le nbre de points initialisé à zéro
  #if OLED
    u8g2.clearBuffer();
    u8g2.setCursor(0,10);  
    u8g2.print("Dist=");
    u8g2.print(dist); 
    u8g2.setCursor(0,22);
    u8g2.print("discent= ");
    u8g2.print(discent);     
    u8g2.setCursor(0,34);
    u8g2.print("xbal= ");
    u8g2.print(xbal);
    u8g2.setCursor(0,46);  
    u8g2.print("ybal= ");
    u8g2.print(xbal);
    u8g2.setCursor(0,64);  
    u8g2.print("point= ");
    u8g2.print(point);       
    u8g2.sendBuffer(); 
  #endif      
} 
//__________________________________________________________________
 
void trt_command() {
  while(SerialBT.available() > 0) {
    cmd = SerialBT.readString();
  } 
  if (digitalRead(fixcib) == LOW) {   // vérifie si cible bien fixée au boitier à chaque lancé de balle      
    if(cmd != "") {
      switch (cmd[0]) {
       case 'R': 
          //Lidar.sendCommand(SOFT_RESET,0);
          //Lidar.sendCommand(SET_FRAME_RATE,FRAME_1000); // nombre de mesures du Lidar par seconde                        
          detectballe();               // lancer détection de la position de la balle
          if (coderr == true) {
          SerialBT.println("E;D;0");       // pour signaler erreur détection balle 
         }
         else {    
           // Envoi bluetooth des mesures
           SerialBT.print(dg);
           SerialBT.print(";");
           SerialBT.print(hb);
           SerialBT.print(";");
           SerialBT.println(point);      
         }
         break;   
       case 'S':
         for (i=0; i<3; i++) {
          sonner(2500,500);
          delay(150);
          sonner(1800,200);
          delay(300);
         }
         break;  
       case 'C':
         calibrage();                  // calibrage par rapport au repère 
         if (coderr == true) {
           SerialBT.println("E;C;0");        // pour signaler erreur de calibrage
           for (i = 0; i < 5; i++) {
             sonner(2000,200);         // alerte sonore
             delay(300);
           }
         }
         else {
          // Envoi bluetooth calibrage ok 
           SerialBT.println("O;K;0");
         }
         break;  
     }
     cmd="";                      // vider la chaine de commande
    }
  }
  else {
    SerialBT.println("E;F;0");         // pour signaler mauvaise fixation de la cible
    for (i = 0; i < 5; i++) {
      sonner(2000,200);          // alerte sonore
      delay(300);
    }
  }
}
//__________________________________________________________________

void setup() {
  Serial.begin(115200);
  EasyBuzzer.update();         // correction bug dans lib EasyBuzzer
  distmin = 10;                // distance mini de détection de la balle (propriété du lidar)
  vadimax = (sqrt(pow(xmax, 2) + pow(ymax, 2)) - xmax) / 256;  // variation de dist.maxi pour un pas d'angle du lidar (long diagonale - long coté / nb pas)
  xcentre = xmax / 2;          // calcul position du centre de la cible
  ycentre = xcentre;
  largano = xcentre / 4;       // calcul largeur d'un anneau de la cible (12,5 cm pour une cible de 100 cm de côté)
  #if OLED
    u8g2.begin();
    u8g2.clearBuffer();                  // effacer mémoire Oled  
    u8g2.setFont(u8g2_font_courB10_tf);  // choix de la police (CF: https://github.com/olikraus/u8g2/wiki/fntlistall)
    u8g2.drawStr(20,10,"CeCible v1");    // Afficher le nom du programme    
    u8g2.setCursor(0,22);
    u8g2.print("Cible ");     
    u8g2.print(int(xmax));
    u8g2.print("x");
    u8g2.print(int(ymax));
    u8g2.setCursor(0,34);
    u8g2.print("D.min = ");     
    u8g2.print(int(distmin));
    u8g2.setCursor(0,46);
    u8g2.print("D.max = ");     
    u8g2.print(int(distmax));
    u8g2.setCursor(0,58);
    u8g2.print("Larg Ann.= ");     
    u8g2.print(int(largano);
    u8g2.sendBuffer();             // envoi du buffer Oled sur écran
  #endif
  #if MOSER 
    Serial.print("Cible ");     
    Serial.print(int(xmax));
    Serial.print("   x");
    Serial.print(int(ymax));
    Serial.print("   D.min = ");     
    Serial.print(int(distmin));
    Serial.print("   D.max = ");     
    Serial.print(int(distmax));
    Serial.print("   Larg Ann.= ");     
    Serial.println(int(largano));
  #endif
  lidar.begin (115200,SERIAL_8N1,16,17) ;   // port série 2: Lidar sur pins 16 et 17
  SerialBT.begin("ESP32");    // Initialisation du bluetooth classique
  cmd.reserve(64);            // réserver 64 octets pour réception message bluetooth
  moteur.connectToPins(mot1,mot2,mot3,mot4);
  moteur.disableMotor();                 // mettre moteur à l'arrêt au démarrage
  moteur.setSpeedInStepsPerSecond(500);
  moteur.setAccelerationInStepsPerSecondPerSecond(1000);
  radstep = (2 * PI) / 2048;                    // calcul de l'angle d'un pas en radians
  margestep = abs((2048 / 360) * margedeg);    // cacul marge de balayage en nb de pas (valeur absolue)
  stepcal = ((2048 / 360) * angdep) * -1;      // calcul angle de calage par rapport au côté droit en nb de pas (nb négatif)
  ledcSetup(0, 5000, 8);           // DAC channel 0 at 5kHz on 8 bits
  ledcAttachPin(ledctl, 0);        // Channel 0 pour led de controle
  ledcWrite(0, 0);                // led de controle off
  EasyBuzzer.setPin(13);       // pin 13 : Pin buzzer
  pinMode(fdcg,INPUT_PULLUP);  // détecteur fin de course gauche à 5v au repos
  pinMode(fdcd,INPUT_PULLUP);  // détecteur fin de course droit à 5v au repos
  pinMode(lopt,INPUT_PULLUP);  // détecteur optique de positionnement (à 0v quand pas de détection)
  pinMode(fixcib,INPUT_PULLUP); // détecteur fixation cible (à 0v si bien fixée)
  sonner(1200,100);
}
//__________________________________________________________________

void loop() {  
  trt_command();               // traitement des messages bluetooth reçus  
}
//______________________FIN PROGRAMME____________________________________________
